# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Callable, Generic, TypeVar

from apgorm.field import Field
from apgorm.undefined import UNDEF

_T = TypeVar("_T", covariant=True)
_S = TypeVar("_S", bound="SqlType", covariant=True)


class SqlType(Generic[_T]):
    """Base type for all SQL types."""

    _sql: str
    """The raw sql for the type."""

    def field(
        self: _S,
        default: _T | UNDEF = UNDEF.UNDEF,
        default_factory: Callable[[], _T] | None = None,
        use_repr: bool = True,
    ) -> Field[_S, _T]:
        """Generate a field using this type."""

        return Field(
            sql_type=self,
            default=default,
            default_factory=default_factory,
            not_null=True,
            use_repr=use_repr,
        )

    def nullablefield(
        self: _S,
        default: _T | None | UNDEF = UNDEF.UNDEF,
        default_factory: Callable[[], _T] | None = None,
        use_repr: bool = True,
    ) -> Field[_S, _T | None]:
        """Generate a nullable field using this type."""

        return Field(
            sql_type=self,
            default=default,
            default_factory=default_factory,
            not_null=False,
            use_repr=use_repr,
        )



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import datetime
from enum import Enum

from .base_type import SqlType


class Timestamp(SqlType[datetime.datetime]):
    """Date and time without time zone.

    Args:
        precision (int, optional): Precision (0-6). Defaults to None.

    https://www.postgresql.org/docs/14/datatype-datetime.html
    """

    def __init__(self, precision: int | None = None) -> None:
        self._precision = precision
        self._sql = "TIMESTAMP"
        if precision is not None:
            self._sql += f"({precision})"

    @property
    def precision(self) -> int | None:
        """The precision of the timestamp.

        Returns:
            int | None
        """

        return self._precision


class TimestampTZ(SqlType[datetime.datetime]):
    """Date and time with time zone.

    Args:
        precision (int, optional): Precision (0-6). Defaults to None.

    https://www.postgresql.org/docs/14/datatype-datetime.html
    """

    def __init__(self, precision: int | None = None) -> None:
        self._precision = precision
        self._sql = "TIMESTAMPTZ"
        if precision is not None:
            self._sql += f"({precision})"

    @property
    def precision(self) -> int | None:
        """The precision of the timestamp.

        Returns:
            int | None
        """

        return self._precision


class Time(SqlType[datetime.time]):
    """Time of day with no date or time zone.

    Args:
        precision (int, optional): Precision (0-6). Defaults to None.

    https://www.postgresql.org/docs/14/datatype-datetime.html
    """

    def __init__(self, precision: int | None = None) -> None:
        self._precision = precision
        self._sql = "TIME"
        if precision is not None:
            self._sql += f"({precision})"

    @property
    def precision(self) -> int | None:
        """The precision of the time.

        Returns:
            int | None
        """

        return self._precision


class TimeTZ(SqlType[datetime.time]):
    """Time of day with time zone (no date).

    Args:
        precision (int, optional): Precision (0-6). Defaults to None.

    https://www.postgresql.org/docs/14/datatype-datetime.html
    """

    def __init__(self, precision: int | None = None) -> None:
        self._precision = precision
        self._sql = "TIMETZ"
        if precision is not None:
            self._sql += f"({precision})"

    @property
    def precision(self) -> int | None:
        """The precsion of the time.

        Returns:
            int | None
        """

        return self._precision


class Date(SqlType[datetime.date]):
    """Date (no time of day).

    https://www.postgresql.org/docs/14/datatype-datetime.html
    """

    _sql = "DATE"


class IntervalField(Enum):
    YEAR = "YEAR"
    MONTH = "MONTH"
    DAY = "DAY"
    HOUR = "HOUR"
    MINUTE = "MINUTE"
    SECOND = "SECOND"
    YEAR_TO_MONTH = "YEAR TO MONTH"
    DAY_TO_HOUR = "DAY TO HOUR"
    DAY_TO_SECOND = "DAY TO SECOND"
    HOUR_TO_MINUTE = "HOUR TO MINUTE"
    HOUR_TO_SECOND = "HOUR TO SECOND"
    MINUTE_TO_SECOND = "MINUTE TO SECOND"


class Interval(SqlType[datetime.timedelta]):
    """Interval type. Essentially a time range (5 years or 2 minutes) but not
    an actual date or time (05-9-2005).

    Args:
        interval_field (IntervalField, optional): Restrict the fields of
        the interval. Defaults to None.
        precision (int, optional): Precision (0-6). Defaults to None.

    https://www.postgresql.org/docs/14/datatype-datetime.html
    """

    def __init__(
        self,
        interval_field: IntervalField | None = None,
        precision: int | None = None,
    ) -> None:
        self._interval_field = interval_field
        self._precision = precision

        self._sql = "INTERVAL"
        if interval_field is not None:
            self._sql += " " + interval_field.value
        if precision is not None:
            self._sql += f"({precision})"

    @property
    def interval_field(self) -> IntervalField | None:
        """The allowed fields for the interval.

        Returns:
            IntervalField | None
        """

        return self._interval_field

    @property
    def precision(self) -> int | None:
        """The precision of the interval.

        Returns:
            int | None
        """

        return self._precision



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from enum import Enum
from typing import Sequence
from typing import cast as typingcast

from apgorm.exceptions import BadArgument
from apgorm.field import BaseField
from apgorm.sql.sql import Block, join, r

from .constraint import Constraint


class ForeignKeyAction(Enum):
    """Action for ON UPDATE or ON DELETE of ForeignKey."""

    CASCADE = "CASCADE"
    """Carry the changes"""
    RESTRICT = "RESTRICT"
    """Prevent the changes"""
    NO_ACTION = "NO ACTION"
    """Do nothing"""


class ForeignKey(Constraint):
    def __init__(
        self,
        fields: Sequence[Block | BaseField | str] | Block | BaseField | str,
        ref_fields: Sequence[Block | BaseField | str]
        | Block
        | BaseField
        | str,
        ref_table: Block | str | None = None,
        match_full: bool = False,
        on_delete: ForeignKeyAction = ForeignKeyAction.CASCADE,
        on_update: ForeignKeyAction = ForeignKeyAction.CASCADE,
    ) -> None:
        """Specify a ForeignKey constraint for a table.

        Args:
            fields (Sequence[Block | BaseField]): A list of fields or raw
            field names on the current table.
            ref_fields (Sequence[Block | BaseField]): A list of fields or raw
            field names on the referenced table.
            ref_table (Block, optional): If all of `ref_fields` are `Block`,
            specify the raw tablename of the referenced table. Defaults to
            None.
            match_full (bool, optional): Whether or not a full match is
            required. If False, MATCH SIMPLE is used instead. Defaults to
            False.
            on_delete (Action, optional): The action to perform if the
            referenced row is deleted. Defaults to Action.CASCADE.
            on_update (Action, optional): The action to perform if the
            referenced row is updated. Defaults to Action.CASCADE.

        Raises:
            BadArgument: Bad arguments were sent to ForeignKey.
        """

        self.fields: Sequence[BaseField | Block] = [
            r(f) if isinstance(f, str) else f
            for f in (fields if isinstance(fields, Sequence) else [fields])
        ]
        self.ref_fields: Sequence[BaseField | Block] = [
            r(f) if isinstance(f, str) else f
            for f in (
                ref_fields
                if isinstance(ref_fields, Sequence)
                else [ref_fields]
            )
        ]
        self.ref_table = (
            r(ref_table) if isinstance(ref_table, str) else ref_table
        )
        self.match_full = match_full
        self.on_delete = on_delete
        self.on_update = on_update

        if len(self.ref_fields) != len(self.fields):
            raise BadArgument(
                "Must have same number of fields and ref_fields."
            )

        if len(self.fields) == 0:
            raise BadArgument("Must specify at least on field and ref_field.")

    def _creation_sql(self) -> Block:
        ref_table: Block
        ref_fields: list[Block] = []

        if (
            len(
                {
                    f.model.tablename
                    for f in self.ref_fields
                    if isinstance(f, BaseField)
                }
            )
            > 1
        ):
            raise BadArgument(
                "All fields in ref_fields must be of the same table."
            )

        if self.ref_table is None:
            _ref_fields = self.ref_fields
            if not all([isinstance(f, BaseField) for f in _ref_fields]):
                raise BadArgument(
                    "ref_fields must either all be BaseFields or "
                    "ref_table must be specified."
                )
            _ref_fields = typingcast(Sequence[BaseField], _ref_fields)

            ref_table = r(_ref_fields[0].model.tablename)

        else:
            ref_table = self.ref_table

        ref_fields = [
            r(f.name) if isinstance(f, BaseField) else f
            for f in self.ref_fields
        ]
        fields = [
            r(f.name) if isinstance(f, BaseField) else f for f in self.fields
        ]

        return Block(
            r("CONSTRAINT"),
            r(self.name),
            r("FOREIGN KEY ("),
            join(r(","), *fields),
            r(") REFERENCES"),
            ref_table,
            r("("),
            join(r(","), *ref_fields),
            (r(") MATCH FULL") if self.match_full else r(") MATCH SIMPLE")),
            r(f"ON DELETE {self.on_delete.value}"),
            r(f"ON UPDATE {self.on_update.value}"),
            wrap=True,
        )



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, Sequence, Type, overload

from apgorm.field import BaseField
from apgorm.sql.sql import SQL, Block, join, r, wrap
from apgorm.types.boolean import Bool
from apgorm.undefined import UNDEF

if TYPE_CHECKING:  # pragma: no cover
    from apgorm.model import Model


@overload
def select(
    *,
    from_: Model | Type[Model] | Block,
    fields: Sequence[BaseField | Block] | None = ...,
    where: Block[Bool] | None = ...,
    order_by: SQL | UNDEF = ...,
    reverse: bool = ...,
    limit: int | None = ...,
) -> Block:
    ...


@overload
def select(
    *,
    from_: Model | Type[Model] | Block,
    count: Literal[True] = True,
    where: Block[Bool] | None = ...,
    limit: int | None = ...,
) -> Block:
    ...


def select(
    *,
    from_: Model | Type[Model] | Block,
    fields: Sequence[BaseField | Block] | None = None,
    count: bool = False,
    where: Block[Bool] | None = None,
    order_by: SQL | UNDEF = UNDEF.UNDEF,
    reverse: bool = False,
    limit: int | None = None,
) -> Block[Any]:
    sql = Block[Any](r("SELECT"))

    if count:
        sql += Block(r("COUNT(1)"))
    elif fields is not None:
        sql += join(r(","), *fields, wrap=True)
    else:
        sql += Block(r("*"))

    tablename = from_ if isinstance(from_, Block) else r(from_.tablename)
    sql += Block(r("FROM"), tablename)

    if where is not None:
        sql += Block(r("WHERE"), wrap(where))

    if order_by is not UNDEF.UNDEF:
        sql += Block(r("ORDER BY"), order_by)
        sql += r("DESC" if reverse else "ASC")

    if limit is not None:
        sql += Block(r("LIMIT"), r(str(limit)))

    return wrap(sql)


def delete(
    from_: Model | Type[Model] | Block,
    where: Block[Bool] | None = None,
    return_fields: Sequence[BaseField | Block] | None = None,
) -> Block:
    tablename = from_ if isinstance(from_, Block) else r(from_.tablename)
    sql = Block[Any](r("DELETE FROM"), tablename)
    if where is not None:
        sql += Block(r("WHERE"), where)
    if return_fields is not None:
        sql += Block(
            r("RETURNING"),
            join(r(","), *return_fields),
        )
    return wrap(sql)


def update(
    table: Model | Type[Model] | Block,
    values: dict[SQL, SQL],
    where: Block[Bool] | None = None,
    return_fields: Sequence[BaseField | Block] | None = None,
) -> Block:
    tablename = table if isinstance(table, Block) else r(table.tablename)
    sql = Block[Any](r("UPDATE"), tablename, r("SET"))

    set_logic: list[Block] = []
    for key, value in values.items():
        set_logic.append(Block(key, r("="), value))

    sql += join(r(","), *set_logic)

    if where is not None:
        sql += Block(r("WHERE"), where)

    if return_fields is not None:
        sql += Block(
            r("RETURNING"),
            join(r(","), *return_fields),
        )

    return wrap(sql)


def insert(
    into: Model | Type[Model] | Block,
    fields: Sequence[BaseField | Block],
    values: Sequence[SQL],
    return_fields: Sequence[BaseField | Block] | None = None,
) -> Block[Any]:
    tablename = into if isinstance(into, Block) else r(into.tablename)

    sql = Block[Any](r("INSERT INTO"), tablename)
    if len(fields) > 0:
        sql += join(r(","), *fields, wrap=True)

    if len(values) > 0:
        sql += Block[Any](r("VALUES"), join(r(","), *values, wrap=True))
    else:
        sql += Block[Any](r("DEFAULT VALUES"))

    if return_fields is not None:
        sql += r("RETURNING")
        if isinstance(return_fields, (BaseField, Block)):
            sql += return_fields
        else:
            sql += join(r(","), *return_fields)

    return wrap(sql)



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING

from apgorm.sql.sql import Block, join, r

from .constraint import Constraint

if TYPE_CHECKING:  # pragma: no cover
    from apgorm.field import BaseField


class PrimaryKey(Constraint):
    def __init__(self, *fields: BaseField | Block | str) -> None:
        """Do not use. Specify primary keys like this instead:

        ```
        class Users(Model):
            username = VarChar(32).field()

            primary_key = (username,)
        ```
        """

        self.fields = fields

    def _creation_sql(self) -> Block:
        fields = [
            r(f)
            if isinstance(f, str)
            else f
            if isinstance(f, Block)
            else r(f.name)
            for f in self.fields
        ]
        return Block(
            r("CONSTRAINT"),
            r(self.name),
            r("PRIMARY KEY ("),
            join(r(","), *fields),
            r(")"),
            wrap=True,
        )



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections import UserString
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Type,
    TypeVar,
    Union,
    overload,
)

if TYPE_CHECKING:  # pragma: no cover
    from apgorm.field import BaseField
    from apgorm.types.base_type import SqlType
    from apgorm.types.boolean import Bool
    from apgorm.types.numeric import Int  # noqa

_T = TypeVar("_T", covariant=True)
_T2 = TypeVar("_T2")
_SQLT_CO = TypeVar("_SQLT_CO", bound="SqlType", covariant=True)
_SQLT = TypeVar("_SQLT", bound="SqlType")
SQL = Union[
    "BaseField[SqlType[_T], _T, Any]",
    "Block[SqlType[_T]]",
    "Parameter[_T]",
    _T,
]
"""
A type alias that allows for any field, raw sql, parameter.

If used like `SQL[int]`, then it must have a python type of int.
For example:

```
integer: SQL[int] = sql(1)  # works
integer: SQL[int] = sql(1).cast(BigInt())  # works
integer: SQL[int] = sql("hi")  # fails
"""

CASTED = Union[
    "BaseField[_SQLT, Any, Any]",
    "Block[_SQLT]",
]
"""
A type alias that allows for any field or raw sql with a specified
SQL type. For example:

```
bigint: CASTED[BigInt] = sql(1).cast(BigInt())  # works
bigint: CASTED[BigInt] = sql(1)  # fails
bigint: CASTED[BigInt] = sql(1).cast(SmallInt())  # fails
```
"""


@overload
def sql(piece: CASTED[_SQLT_CO], /, *, wrap: bool = ...) -> Block[_SQLT_CO]:
    ...


@overload
def sql(piece: SQL[_T2], /, *, wrap: bool = ...) -> Block[SqlType[_T2]]:
    ...


@overload
def sql(*pieces: SQL, wrap: bool = ...) -> Block:
    ...


def sql(*pieces: SQL, wrap: bool = False) -> Block:
    """Convenience function to wrap content in a Block.

    It is safe to pass user input to this function.

    Example:
    ```
    sql(
        r("SELECT * FROM mytable WHERE id="),
        5,
        r("AND othercol="),
        "hello",
    ).render()
    # "SELECT * FROM mytable WHERE id=$1 AND othercol=$2", [5, "hello"]
    ```

    Returns:
        Block: The pieces wrapped in a Block.
    """

    return Block(*pieces, wrap=wrap)


def wrap(*pieces: SQL) -> Block:
    """Return the SQL pieces as a Block, wrapping them in parantheses."""

    return sql(*pieces, wrap=True)


def join(joiner: SQL, *values: SQL, wrap: bool = False) -> Block:
    """SQL version of "delim".join(values). For example:

    ```
    join(r(","), value1, value2, value3)
    ```

    Args:
        joiner (SQL): The joiner or deliminator.

    Kwargs:
        wrap (bool): Whether or not to return the result wrapped. Defaults to
        False.
    """

    new_values: list[SQL] = []
    for x, v in enumerate(values):
        new_values.append(v)
        if x != len(values) - 1:
            new_values.append(joiner)
    return sql(*new_values, wrap=wrap)


def or_(*pieces: SQL[bool]) -> Block[Bool]:
    """Shortcut for `join(r("OR"), value1, value2, ...)`."""

    return join(r("OR"), *pieces, wrap=True)


def and_(*pieces: SQL[bool]) -> Block[Bool]:
    """Shortcut for `join(r("AND"), value1, value2, ...)`."""

    return join(r("AND"), *pieces, wrap=True)


def r(string: str) -> Block:
    """Treat the string as raw SQL and return a Block.

    Never, ever, ever wrap user input in `r()`. A good rule of thumb is that
    if you see a call to `r()`, the input should be text in quotes, not a
    variable.

    ```
    r(somevarthatmightbeuserinput)  # bad
    r("some text")  # good
    ```
    """

    return sql(Raw(string))


class Raw(UserString):
    pass


class Parameter(Generic[_T]):
    def __init__(self, value: _T) -> None:
        self.value = value


class _Op(Generic[_SQLT]):
    def __init__(self, op: str) -> None:
        self.op = op

    def __get__(
        self, inst: object, cls: Type[object]
    ) -> Callable[[SQL], Block[_SQLT]]:
        def operator(other: SQL) -> Block[_SQLT]:
            assert isinstance(inst, Comparable)
            return wrap(inst._get_block(), r(self.op), other)

        return operator


class _Func(Generic[_SQLT]):
    def __init__(self, func: str, rside: bool = False) -> None:
        self.func = func
        self.rside = rside

    def __get__(self, inst: object, cls: Type[object]) -> Block[_SQLT]:
        assert isinstance(inst, Comparable)
        if self.rside:
            return wrap(wrap(inst._get_block()), r(self.func))
        return wrap(r(self.func), wrap(inst._get_block()))


class Comparable:
    def _get_block(self) -> Block:
        raise NotImplementedError  # pragma: no cover

    def cast(self, type_: _SQLT) -> Block[_SQLT]:
        return wrap(self._get_block(), r("::"), r(type_._sql))

    # comparison
    is_null = _Func["Bool"]("IS NULL", True)
    is_true = _Func["Bool"]("IS TRUE", True)
    is_false = _Func["Bool"]("IS FALSE", True)
    num_nulls = _Func["Int"]("NUM_NULLS")
    num_nonnulls = _Func["Int"]("NUM_NONNULLS")
    not_ = _Func["Bool"]("NOT")

    eq = _Op["Bool"]("=")
    neq = _Op["Bool"]("!=")
    lt = _Op["Bool"]("<")
    gt = _Op["Bool"](">")
    lteq = _Op["Bool"]("<=")
    gteq = _Op["Bool"](">=")

    any = _Func["Any"]("ANY")
    all = _Func["Any"]("ALL")


class Block(Comparable, Generic[_SQLT_CO]):
    """Represents a list of raw sql and parameters."""

    def __init__(self, *pieces: SQL | Raw, wrap: bool = False) -> None:
        """Create a Block. You may find it more convienient to use the `sql()`
        helper function (or `wrap(sql())` if you want the result to be wrapped
        in "( )").

        Kwargs:
            wrap (bool): Whether or not to add `( )` around the result.

        Example:
        ```
        sql = Block(r("SELECT"), "Hello, World", r(","), 17)
        sql.render()  # "SELECT $1, $2", ["Hello, World", 17]
        ```
        """

        self._pieces: list[Raw | Parameter] = []

        if len(pieces) == 1 and isinstance(pieces[0], Block):
            block = pieces[0]
            assert isinstance(block, Block)
            self._wrap: bool = block._wrap or wrap
            self._pieces = block._pieces

        else:
            self._wrap = wrap
            for p in pieces:
                if isinstance(p, Comparable):
                    p = p._get_block()
                if isinstance(p, Block):
                    self._pieces.extend(p.get_pieces())
                elif isinstance(p, (Raw, Parameter)):
                    self._pieces.append(p)
                else:
                    self._pieces.append(Parameter(p))

    def render(self) -> tuple[str, list[Any]]:
        """Return the rendered result of the Block.

        Returns:
            tuple[str, list[Any]]: The raw SQL (as a string) and the list of
            parameters.
        """

        return Renderer().render(self)

    def render_no_params(self) -> str:
        """Convenience function to get the SQL but not the parameters.

        Although it is technically easier to write `.render()[0]`, this is
        cleaner and more obvious to those reading your code.

        Returns:
            str: The sql.
        """

        return self.render()[0]

    def get_pieces(
        self, force_wrap: bool | None = None
    ) -> list[Raw | Parameter]:
        wrap = self._wrap if force_wrap is None else force_wrap
        if wrap:
            return [Raw("("), *self._pieces, Raw(")")]
        return self._pieces

    def _get_block(self) -> Block:
        return self

    def __iadd__(self, other: object):
        if isinstance(other, Block):
            self._pieces.extend(other.get_pieces())
        elif isinstance(other, Parameter):
            self._pieces.append(other)
        else:
            raise TypeError(f"Unsupported type {type(other)}")

        return self


class Renderer:
    def __init__(self) -> None:
        self._curr_value_id: int = 0

    @property
    def _next_value_id(self) -> int:
        self._curr_value_id += 1
        return self._curr_value_id

    def render(self, sql: Block) -> tuple[str, list[Any]]:
        sql_pieces: list[str] = []
        params: list[Any] = []

        for piece in sql.get_pieces(force_wrap=False):
            if isinstance(piece, Raw):
                sql_pieces.append(str(piece))
            else:
                sql_pieces.append(f"${self._next_value_id}")
                params.append(piece.value)

        return " ".join(sql_pieces), params



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations, print_function

from typing import TYPE_CHECKING, Any, Type, TypeVar

from .constraints.check import Check
from .constraints.constraint import Constraint
from .constraints.exclude import Exclude
from .constraints.foreign_key import ForeignKey
from .constraints.primary_key import PrimaryKey
from .constraints.unique import Unique
from .exceptions import ModelNotFound, SpecifiedPrimaryKey
from .field import BaseField, ConverterField
from .manytomany import ManyToMany
from .migrations.describe import DescribeConstraint, DescribeTable
from .sql.query_builder import (
    DeleteQueryBuilder,
    FetchQueryBuilder,
    InsertQueryBuilder,
    UpdateQueryBuilder,
)
from .undefined import UNDEF

if TYPE_CHECKING:  # pragma: no cover
    from .connection import Connection
    from .database import Database


_SELF = TypeVar("_SELF", bound="Model")


class Model:
    """Base class for all models. To create a new model, subclass this class
    and add it to your database like this:

    ```
    class User(Model):
        username = VarChar(32).field()
        ...

        primary_key = (username,)

    ...

    class MyDatabase(Database):
        users = User
        ...
    ```

    Use Model() to create an instance of the model. There are two reasons you
    might do this:
        1. You used `Database.fetch...` directly instead of `Model.fetch`, and
        you want to convert the result to a model.
        2. You want to create a new model (`Model(**values).create()`).

    If you wish to fetch an existing model, please use `Model.fetch` or
    `Model.fetch_query`.
    """

    all_fields: dict[str, BaseField]
    all_constraints: dict[str, Constraint]
    _all_mtm: dict[str, ManyToMany]

    tablename: str  # populated by Database
    """The name of the table, populated by Database."""
    database: Database  # populated by Database
    """The database instance, populated by Database."""

    primary_key: tuple[BaseField, ...]
    """The primary key for the model. All models MUST have a primary key."""

    def __init__(self, override_validator: bool = False, /, **values) -> None:
        copied_fields: dict[str, BaseField] = {}

        for f in self.all_fields.values():
            f = f.copy()
            copied_fields[f.name] = f
            setattr(self, f.name, f)

            convert: bool = True
            value = values.get(f.name, UNDEF.UNDEF)
            if value is UNDEF.UNDEF:
                d = f._get_default()
                convert = False
                if d is not UNDEF.UNDEF:
                    value = d
                else:
                    continue  # pragma: no cover  # (pytest bug)
            if isinstance(f, ConverterField) and convert:
                if not override_validator:
                    f._validate(value)
                f._value = f.converter.to_stored(value)
            else:
                if not override_validator:
                    if isinstance(f, ConverterField):
                        f._validate(f.converter.from_stored(value))
                    else:
                        f._validate(value)
                f._value = value

        # carry the copies of the fields over to primary_key so that
        # Model.field is Model.primary_key[index of that field]
        self.primary_key = tuple(
            [copied_fields[f.name] for f in self.primary_key]
        )

        self.all_fields = copied_fields

        for name, mtm in self._all_mtm.items():
            mtm = mtm._generate_mtm(self)
            setattr(self, name, mtm)

    async def delete(self: _SELF, con: Connection | None = None) -> _SELF:
        """Delete the model. Does not update the values of this model,
        but the returned model will have updated values.

        Returns:
            Model: The deleted model (with updated values).
        """

        deleted = (
            await self.delete_query(con=con)
            .where(**(f := self._pk_fields()))
            .execute()
        )
        if len(deleted) == 0:
            raise ModelNotFound(self.__class__, f)
        return deleted[0]

    async def save(self, con: Connection | None = None) -> None:
        """Save any changed fields of the model. Updates the values
        of this model."""

        changed_fields = self._get_changed_fields()
        if len(changed_fields) == 0:
            return
        q = self.update_query(con=con).where(**self._pk_fields())
        q.set(**{f.name: f._value for f in changed_fields})
        result = await q.execute()
        self._update_from_model(result[0])
        self._set_saved()

    async def create(self: _SELF, con: Connection | None = None) -> _SELF:
        """Insert the model into the database. Updates the values on this
        model.

        Returns:
            Model: The model that was inserted. Useful if you want to write
            `model = await Model(...).create()`
        """

        q = self.insert_query(con=con)
        q.set(
            **{
                f.name: f._value
                for f in self.all_fields.values()
                if f._value is not UNDEF.UNDEF
            }
        )
        self._update_from_model(await q.execute())

        return self

    async def refetch(self: _SELF, con: Connection | None = None) -> None:
        """Updates the model instance by fetching changed values from the
        database."""

        res = await self.fetch(con, **self._pk_fields())
        self._update_from_model(res)

    @classmethod
    async def exists(
        cls: Type[_SELF], con: Connection | None = None, /, **values
    ) -> _SELF | None:
        """Check if a model with the given values exists.

        ```
        assert (await User.exists(nickname=None)) is not None
        ```

        This method can also be used to get the `model or None`, whereas
        `.fetch()` will raise an exception if the model does not exist.

        Returns:
            Model | None: The model if it existed, otherwise None.
        """

        try:
            return await cls.fetch(con, **values)
        except ModelNotFound:
            return None

    @classmethod
    async def fetch(
        cls: Type[_SELF],
        con: Connection | None = None,
        /,
        **values,
    ) -> _SELF:
        """Fetch an exiting model from the database.

        Example:
        ```
        user = await User.fetch(username="Circuit")
        ```

        Raises:
            ModelNotFound: No model for the given parameters were found.

        Returns:
            Model: The model.
        """

        res = await cls.fetch_query(con=con).where(**values).fetchone()
        if res is None:
            raise ModelNotFound(cls, values)
        return res

    @classmethod
    async def count(
        cls: Type[_SELF],
        con: Connection | None = None,
        /,
        **values,
    ) -> int:
        """Get the number of rows matching these params.

        Returns:
            int: The count.
        """

        return await cls.fetch_query(con=con).where(**values).count()

    @classmethod
    def fetch_query(
        cls: Type[_SELF], con: Connection | None = None
    ) -> FetchQueryBuilder[_SELF]:
        """Returns a FetchQueryBuilder."""

        return FetchQueryBuilder(model=cls, con=con)

    @classmethod
    def delete_query(
        cls: Type[_SELF], con: Connection | None = None
    ) -> DeleteQueryBuilder[_SELF]:
        """Returns a DeleteQueryBuilder."""

        return DeleteQueryBuilder(model=cls, con=con)

    @classmethod
    def update_query(
        cls: Type[_SELF], con: Connection | None = None
    ) -> UpdateQueryBuilder[_SELF]:
        """Returns an UpdateQueryBuilder."""

        return UpdateQueryBuilder(model=cls, con=con)

    @classmethod
    def insert_query(
        cls: Type[_SELF], con: Connection | None = None
    ) -> InsertQueryBuilder[_SELF]:
        """Returns an InsertQueryBuilder."""

        return InsertQueryBuilder(model=cls, con=con)

    def _update_from_model(self: _SELF, updated: _SELF) -> None:
        for f in updated.all_fields.values():
            self.all_fields[f.name]._value = f._value

    @classmethod
    def _primary_key(cls) -> PrimaryKey:
        pk = PrimaryKey(*cls.primary_key)
        pk.name = (
            f"_{cls.tablename}_"
            + "{}".format("_".join([f.name for f in cls.primary_key]))
            + "_primary_key"
        )
        return pk

    @classmethod
    def _describe(cls) -> DescribeTable:
        unique: list[DescribeConstraint] = []
        check: list[DescribeConstraint] = []
        fk: list[DescribeConstraint] = []
        exclude: list[DescribeConstraint] = []
        for c in cls.all_constraints.values():
            if isinstance(c, Check):
                check.append(c._describe())
            elif isinstance(c, ForeignKey):
                fk.append(c._describe())
            elif isinstance(c, Unique):
                unique.append(c._describe())
            elif isinstance(c, Exclude):
                exclude.append(c._describe())

        return DescribeTable(
            name=cls.tablename,
            fields=[f._describe() for f in cls.all_fields.values()],
            fk_constraints=fk,
            pk_constraint=cls._primary_key()._describe(),
            unique_constraints=unique,
            check_constraints=check,
            exclude_constraints=exclude,
        )

    @classmethod
    def _special_attrs(
        cls,
    ) -> tuple[dict[str, BaseField], dict[str, Constraint]]:
        fields: dict[str, BaseField] = {}
        constraints: dict[str, Constraint] = {}
        mtm: dict[str, ManyToMany] = {}

        for attr_name in dir(cls):
            try:
                attr = getattr(cls, attr_name)
            except AttributeError:  # pragma: no cover
                continue

            if isinstance(attr, BaseField):
                fields[attr_name] = attr

            elif isinstance(attr, Constraint):
                if isinstance(attr, PrimaryKey):
                    raise SpecifiedPrimaryKey(
                        cls.__name__,
                        [
                            f.name
                            if isinstance(f, BaseField)
                            else f
                            if isinstance(f, str)
                            else f.render_no_params()
                            for f in attr.fields
                        ],
                    )
                constraints[attr_name] = attr

            elif isinstance(attr, ManyToMany):
                mtm[attr_name] = attr

        cls.all_fields = fields
        cls.all_constraints = constraints
        cls._all_mtm = mtm

        return fields, constraints

    def _pk_fields(self) -> dict[str, Any]:
        return {f.name: f.v for f in self.primary_key}

    def _set_saved(self) -> None:
        for f in self.all_fields.values():
            f.changed = False

    def _get_changed_fields(self) -> list[BaseField]:
        return [f for f in self.all_fields.values() if f.changed]

    # magic methods
    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            + " ".join(
                [
                    f"{f.name}:{f.v!r}"
                    for f in self.all_fields.values()
                    if f.use_repr and f._value is not UNDEF.UNDEF
                ]
            )
            + ">"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            raise TypeError(f"Unsupported type {type(other)}.")

        return self._pk_fields() == other._pk_fields()



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import (
    Any,
    Callable,
    Generator,
    Generic,
    Sequence,
    TypeVar,
    overload,
)

_IN = TypeVar("_IN")
_OUT = TypeVar("_OUT")


class LazyList(Generic[_IN, _OUT]):
    """Lazily converts each value of an iterable.

    Incredible useful for casting `list[asyncpg.Record]` to `list[dict]`, and
    then `list[dict]` to `list[Model]`, especially when there are many rows.
    """

    def __init__(
        self,
        data: Sequence[_IN] | LazyList[Any, _IN],
        converter: Callable[[_IN], _OUT],
    ) -> None:
        self._data = data
        self._converter = converter

    def _convert(self, values: Sequence[_IN] | None = None) -> list[_OUT]:
        _values = self._data if not values else values
        return [self._converter(v) for v in _values]

    @overload
    def __getitem__(self, index: int) -> _OUT:
        ...  # pragma: no cover

    @overload
    def __getitem__(self, index: slice) -> LazyList[_IN, _OUT]:
        ...  # pragma: no cover

    def __getitem__(self, index: int | slice) -> LazyList[_IN, _OUT] | _OUT:
        if isinstance(index, int):
            return self._convert([self._data[index]])[0]
        return LazyList(self._data[index], self._converter)

    def __iter__(self) -> Generator[_OUT, None, None]:
        for r in self._data:
            yield self._converter(r)

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        if len(self) == 5:
            ddd = ", {}".format(repr(self[-1]))
        elif len(self) > 5:
            ddd = ", ..., {}".format(repr(self[-1]))
        else:
            ddd = ""
        return "LazyList([{}{}])".format(
            ", ".join([repr(c) for c in list(self[0:5])]),
            ddd,
        )
